# Copyright 2019 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This package contains the modules that make up the
Strawberry Fields application layer.

.. currentmodule:: strawberryfields.apps
.. autosummary::
    :toctree: api

	clique
	data
	plot
	points
	sample
	similarity
	subgraph
	vibronic
"""

import strawberryfields.apps.clique
import strawberryfields.apps.data
import strawberryfields.apps.plot
import strawberryfields.apps.points
import strawberryfields.apps.sample
import strawberryfields.apps.similarity
import strawberryfields.apps.subgraph
import strawberryfields.apps.vibronic
